/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "gatemon_utils.hh"
#include "DVecType.hh"
#include "DaccAPI.hh"
#include "GDSPlot.hh"
#include "Trend.hh"
#include <sstream>

using namespace std;

static const double dinf(3e5);

//======================================  gate condition constructor
gate_cond::gate_cond(void)
   : _bitno(0), _invert(true), _tvec_limit(4096.0), _thrcut(1.0), _debug(0)
{}

//======================================  Collect gate statistics
gate_cond::gate_cond(const std::string& name)
   : _name(name), _bitno(0), _invert(true), _tvec_limit(4096.0), _thrcut(1.0),
     _debug(0)
{}

//======================================  Collect gate statistics
void
gate_cond::collect(DaccAPI& dapi, bool ready)
{
   //-----------------------------------  Track the gate state
   if (!_gatevec.empty()) {
      uint32_t mask = 1 << _bitno;
      const TSeries* bvec = dapi.refData(_gatevec);
      if (!bvec) {
	 ostringstream msg;
	 msg << "gating_monitor: Bit vector channel not found: " << _gatevec;
	 throw runtime_error(msg.str());
      }
      if (_debug) bvec->Dump(cout);
      const DVecType<uint32_t>*
	 dvi = dynamic_cast<const DVecType<uint32_t>*>(bvec->refDVect());
      if (!dvi)  {
	 throw runtime_error("gating_monitor: Unexpected bit vector data type");
      }
      size_t N = dvi->size();
      Time start;
      bool in_gate = false;
      for (size_t i=0; i<N; i++) {
	 bool discr = bool(dvi->refTData()[i] & mask) != _invert;
	 if (!in_gate) {
	    if (discr) {
	       start = bvec->getBinT(i);
	       if (_debug) {
		  cout << "Start " << name() << " segment at " << start.totalS()
		       << endl;
	       }
	    }
	 } else if (!discr) {
	    Time endT(bvec->getBinT(i));
	    _timeline.add_seg(start, endT - start);
	    if (_debug) {
	       cout << "Stop " << name() << " segment at " << endT.totalS()
		    << endl;
	    }
	 }
	 in_gate = discr;
      }
      if (in_gate) {
	 Time endT(bvec->getBinT(N));
	 _timeline.add_seg(start, endT - start);
	 if (_debug) {
	    cout << "Stop " << name() << " segment at " << endT.totalS()
		 << endl;
	 }
      }
   }

   //-----------------------------------  Record any discriminator values
   if (ready && !_thresh.empty()) {
      TSeries tval(*dapi.refData(_thresh));
      if (!_threshval.empty() && _threshval.getEndTime() < tval.getStartTime()){
	 _threshval.extend(tval.getStartTime());
      }
      _threshval.Append(tval);
      Interval lThresh = _threshval.getInterval();
      if (lThresh > _tvec_limit) {
	 _threshval.eraseStart(lThresh - _tvec_limit);
      }
   }
}

//======================================  Plot the gate channel distribution
void
gate_cond::plot_threshold(GDSPlot& plot) const {
   if (_threshval.empty()) return;
   if (!dynamic_cast<const DVectD*>(_threshval.refDVect())) return;
   _plot_parameters.set_parameters(plot);
   plot.histogram(dynamic_cast<const DVectD&>(*_threshval.refDVect()));
   if (_thrcut > 0.0) {
      plot.line(_thrcut, 0.0, _thrcut, dinf);
   }
}

//======================================  Set the gate type name.
void
gate_cond::set_debug(int dl) {
   _debug = dl;
}

//======================================  Set the bit vector channel name
void
gate_cond::set_gatevec(const std::string& bvec, size_t bit) {
   _gatevec = bvec;
   _bitno   = bit;
}

//======================================  Set the gate type name.
void
gate_cond::set_name(const std::string& name) {
   _name = name;
}

//======================================  Set the threshold channel name
void
gate_cond::set_invert(bool inv) {
   _invert = inv;
}

//======================================  Set the plot parameters
void
gate_cond::set_plot(const plot_pars& pp) {
   _plot_parameters = pp;
   _plot_parameters.status(cout);
}

//======================================  Set the threshold channel name
void
gate_cond::set_thrcut(double thr) {
   _thrcut = thr;
}

//======================================  Set the threshold channel name
void
gate_cond::set_thresh(const std::string& thr) {
   _thresh = thr;
}

//======================================  Set the threshold channel name
void
gate_cond::set_tvec_limit(Interval sz) {
   _tvec_limit = sz;
}

//======================================  Round a segment to second boundaries
inline timeline::time_seg
roundup(const timeline::time_seg& s) {
   Time ts(s.start().getS(), 0);
   Time te(s.end());
   if (te.getN() != 0) te = Time(te.getS() + 1, 0);
   return timeline::time_seg(ts, te);
}

//======================================  Segment definition constructor
segment_def::segment_def(const std::string& name,
					 trig::SegAccountant& act)
   : _name(name), _mode(k_idle), _invert(false), _on_time(0), _avg_time(0),
     _fraction(0), _debug(0), _acct(act), _count(0), _total(0)
{}

//======================================  Translate segment_mode tostring
std::string
segment_def::mode_name(segment_mode s_mode) {
   switch (s_mode) {
   case k_idle:            ///< no segments generated
      return "idle";
   case k_active:          ///< segment is on
      return "active";
   case k_long:            ///< segments based on gate length
      return "long";
   case k_fract_real:      ///< segments from fraction of real time
      return "fract_real";
   case k_fract_live:      ///< segments from fraction of live time
      return "fract_live";
   case k_fract_lock:      ///< Fraction of time in single lock, <= avg
      return "fract_lock";
   }
   return "";
}

//======================================  Write rounded segment, keep statistics
void
segment_def::put_segment(const Time& t0, const Time& t1, bool state) {
   if (t0 && t1 > t0) put_segment(timeline::time_seg(t0, t1), state);
}

//======================================  Write rounded segment, keep statistics
void
segment_def::put_segment(const timeline::time_seg& s, bool state) {
   timeline::time_seg seg(s);
   if (state) {
      seg = roundup(seg);
      if (seg.start() > _active) _count++;
      _total  += seg.duration();
      _active  = seg.end();
   }
   //cout << "put_segment: " << name() << " " << seg.start().getS() << " - "
   //     << seg.end().getS() << " state: " << state << endl;
   _acct.set_segment(name(), seg.start(), seg.end(), state);
   _last = seg.end();
}

//======================================  Segment definition constructor
void
segment_def::seg_gen(const timeline& gates,
		     const timeline& live,
		     const Time& t0, Interval dt) {
   //-----------------------------------  Skip ahead if analysis not up to date
   if (_analyzd < t0) {
      if (_last && _last < _analyzd) put_segment(_last, _analyzd, false);
      _last = t0;
   }
   timeline::time_seg delta(t0, dt);
   Time endStride(delta.end());

   //-----------------------------------  Loop over dead segments
   timeline dead = gates & live;
   for (auto ds : (dead & delta)) {
      timeline::time_seg avg_epoch(delta.end() - _avg_time, _avg_time);
      Interval live_time(0);

      switch (_mode) {
      case k_idle:
	 //-----------------------------  No segments generated
	 break;

      case k_active:
	 //-----------------------------  Generate segments for all gates
	 put_segment(_last, roundup(ds)._t0, false); // -ve segs ignored.
	 put_segment(ds);
	 break;

      case k_long: {
	 //-----------------------------  Generate segments for long gates
	 timeline::time_seg nowgate = dead[ds._t0];
	 if (!nowgate) {
            break;
	 } else if (nowgate._dt > _on_time) {
            timeline::time_seg nowseg = roundup(nowgate);
	    put_segment(_last, nowseg._t0, false); // -ve segs ignored
	    if (_active > nowseg._t0) {
               nowseg = timeline::time_seg(_active, nowseg.end());
            }
	    put_segment(nowseg);
	 } else if (nowgate.end() == delta.end()) {
	    endStride = nowgate._t0;
	 }
	 break;
      }

      case k_fract_real:
	 //-----------------------------  Gate lengths > real time fraction
         live_time = _avg_time;
	 break;

      case k_fract_live:
	 //-----------------------------  Gate lengths > live time fraction
	 live_time = live.on_time(avg_epoch);
	 break;

      case k_fract_lock:
	 //-----------------------------  Gate lengths > lock time fraction
	 avg_epoch = live.back();
	 live_time = _avg_time;
	 break;
      }

      //--------------------------------  Check the fractions.
      if (live_time) {
	 double fract = (dead & avg_epoch).on_time() / live_time;
	 if (fract > _fraction) {
	    put_segment(_last, roundup(ds).start(), _last == _active);
	    put_segment(ds, true);
	 }
      }
   }

   //-----------------------------------  Fill in/extend empty segments
   if (_last < endStride) {
      if (!(live & delta).empty()) {
	 double live_time = 0.0;
	 timeline::time_seg avg_epoch(delta.end() - _avg_time, _avg_time);
	 if (_mode == k_fract_real) {
	    live_time = _avg_time;
	 } else  if (_mode == k_fract_live) {
	    live_time = (live & avg_epoch).on_time();
	 } else  if (_mode == k_fract_lock) {
	    avg_epoch = live.back();
	    live_time = _avg_time;
	 }
	 if (live_time) {
	    double fract = (dead & avg_epoch).on_time() / live_time;
	    if (fract > _fraction) put_segment(_last, endStride, true);
	 }
      }
      put_segment(_last, endStride, false);
   }
   _analyzd = delta.end();
}

//======================================   Set averaging time
void
segment_def::set_average_time(Interval avg) {
   _avg_time = avg;
}

//======================================   Set the debug print level
void
segment_def::set_debug(int d) {
   _debug = d;
}

//======================================  Set the deadtime fraction limit
void
segment_def::set_fraction(double frac) {
   _fraction = frac;
}

//======================================   Set gate condition name
void
segment_def::set_gate(const std::string& name) {
   _gate_cond = name;
}

//======================================  Set invert selection
void
segment_def::set_invert(bool inv) {
   _invert = inv;
}

//======================================  Set the segment calculation mode
void
segment_def::set_mode(segment_mode mode){
   _mode=mode;
}

//======================================  Set the mode from a string
bool
segment_def::set_mode(const std::string& s_mode){
   if (s_mode == "idle") {
      set_mode(k_idle);            ///< no segments generated
   } else if (s_mode == "active") {
      set_mode(k_active);          ///< segment is on
   } else if (s_mode == "long") {
      set_mode(k_long);            ///< segments based on gate length
   } else if (s_mode == "fract_real") {
      set_mode(k_fract_real);      ///< segments from fraction of real time
   } else if (s_mode == "fract_live") {
      set_mode(k_fract_live);      ///< segments from fraction of live time
   } else if (s_mode == "fract_lock") {
      set_mode(k_fract_lock);      ///< Fraction of time in single lock, <= avg
   } else {
      std::ostringstream msg;
      msg << "gating_monitor::segment_def::set_mode: Invalid segment mode: "
	  << s_mode << endl;
      return false;
   }
   return true;
}

//======================================  Set the segment name
void
segment_def::set_name(const std::string& name) {
   _name = name;
}
//======================================  Set the maximum gate length
void
segment_def::set_on_time(Interval odt){
   _on_time = odt;
}

//======================================   Trend_def constructor
trend_def::trend_def(const std::string& name,
				     const std::string& gate)
   : _name(name), _gate_cond(gate), _enable_dead(true), _enable_discr(false)
{}

//======================================   Enable dead-time trending
void
trend_def::set_enable_dead(bool enable) {
   _enable_dead = enable;
}

//======================================   Enable discriminant channel trending
void
trend_def::set_enable_discr(bool enable) {
   _enable_discr = enable;
}

//======================================   Enable discriminant channel trending
void
trend_def::set_prefix(const std::string& prefix) {
   _chan_prefix = prefix;
}

//======================================   Enable discriminant channel trending
void
trend_def::trend_setup(Trend& trend, const std::string& chan_pfx) {
   if (_chan_prefix.empty()) _chan_prefix = chan_pfx + "_" + _name;
   if (_enable_dead) {
      string chan = _chan_prefix + "_DEAD";
      cout << "trend_def: add channel: " << chan << endl;
      trend.addChannel(chan);
   }
   if (_enable_discr) {
      string chan = _chan_prefix + "_DISCR";
      cout << "trend_def: add channel: " << chan << endl;
      trend.addChannel(chan);
   }
}

//======================================   Trend_def constructor
void
trend_def::trend_stride(const Time& t0, Interval dt,
			const gate_cond& gc, Trend& trend) {
   if (_enable_dead) {
      string chan = _chan_prefix + "_DEAD";
      timeline dead(gc.time_line() & timeline::time_seg(t0, dt));
      for (auto ec : dead) {
	 trend.trendData(chan, ec.start(), double(ec.duration()));
      }
   }
   if (_enable_discr) {
      string chan = _chan_prefix + "_DISCR";
      timeline dead(gc.time_line() & timeline::time_seg(t0, dt));
      for (auto ec : dead) {
	TSeries val = gc.discriminator().extract(ec.start(), ec.duration());
	if (!val.empty()) trend.trendData(chan, val);
      }
   }
}
