/* -*- mode: c++; c-basic-offset: 3; -*- */
#ifndef TRENDVIEW_HH
#define TRENDVIEW_HH

#include "DatEnv.hh"
#include "MonServer.hh"
#include "FixedLenTS.hh"
#include "Trend.hh"
#include "autopipe.hh"
#include <string>
#include <vector>
#include <iosfwd>

//
//    Data Monitor template class.

/**  @name trendview
  *  \brief Monitor to trend and display filtered data.
  *  The trendview monitor filters one or more input channels and trends and/or
  *  displays the result.
  *
  *  {\bf {\Large Running trendview}}
  *
  *  The trendview command line syntax is:
  *  \begin{verbatim}
      trendview -config <json-file>
     \end{verbatim}
  *
  *  {\bf {\Large Configuration File}}
  *
  *  trendview reads a json configuration file
  *
  *  {\bf {\Large trendview Output}}
  *
  *  {\bf Trend channels}
  *  trendview produces a minute trend of the sigma value called 
  *  #<channel>_sigma#
  *
  *  {\bf Monitor data channels}
  *  trendview publishes a time series history of trended filtered values.
  *  with the name #<channel>_sigma#
  *
  *  \author John Zweizig; Caltech/LIGO
  *  \version 1.0; Last modified July 16, 2014
  */
//{
//}

// classdef
class trendview : public DatEnv, MonServer
{
   class channel_info {
   public:
      channel_info(const std::string& name);
      ~channel_info(void);
      void add_data(const TSeries& ts);
      void book_trend(Trend& trender);
      void book_viewer(MonServer& viewer) const;
      int debug(void) const;
      void display(std::ostream& out) const;
      const TSeries& filtered_data(void);
      void init_data(const std::string& monitor);
      void reset(const Time& t);
      void set_debug(int dbg);
      void set_filter(const std::string& filt);
      void set_trend_channel(const std::string& tchan);
      void set_view_channel(const std::string& vchan);
      void set_view_epoch(Interval epoch);
      void set_view_stride(Interval vstride);
      const std::string& data_channel(void) const;
      const std::string& trend_channel(void) const;
      const std::string& view_channel(void) const;
   private:
      std::string mDataChannel;
      std::string mTrendChannel;
      std::string mViewChannel;
      std::string mFilter;        //  Filter description
      Interval    mSettle;        //  Settling time.
      Interval    mViewEpoch;
      Interval    mViewStride;
      int         mDebug;
      
      TrendAcc    mViewAcc;
      auto_pipe   mPipe;
      TSeries     mFiltered;
      Time        mStartTime;
      Time        mCurrent;
      Time        mAccTime;
      FixedLenTS  mData;
      FixedLenTS  mMinimum;
      FixedLenTS  mMaximum;
      FixedLenTS  mSigma;
   };
   typedef std::vector<channel_info> chan_vect;
  
public:
   /**  trendview monitor class construtor
     */
   trendview(int argc, const char *argv[]);   // Constructor
    
   /**  trendview monitor class construtor
     */
   ~trendview(void);

   /** Configure the monitor from a json file
     * \param file json configuration file.
     */
   void configure(const std::string& file);
    
   /**  Trend the input data on all configured channels and make it available 
     *  for viewing via the dmtviewer mechanism.
     *  \brief Process one strid of raw data.
     */
   void ProcessData(void);
   
   /**  Handle a viewer attention interrupt.
     *  \brief Attention handler.
     */
   void Attention(void);
    
private:
   std::string mServerName;    //  monitor data server name
   std::string mTrendName;     //  trend file name
   Interval    mStep;          //  Time stride.
   Trend       mTrend;         //  Trend writer
   chan_vect   mChannel;       //  Channel vector
};

inline const std::string&
trendview::channel_info::data_channel(void) const {
   return mDataChannel;
}

inline int
trendview::channel_info::debug(void) const {
   return mDebug;
}

inline const TSeries& 
trendview::channel_info::filtered_data(void) {
   return mFiltered;
}

inline const std::string&
trendview::channel_info::trend_channel(void) const {
   return mTrendChannel;
}

inline const std::string& 
trendview::channel_info::view_channel(void) const {
   return mViewChannel;
}

#endif     //  TRENDVIEW_HH
